@@ -1,4 +1,4 @@
-from integration_tests.data_sources import ClickhouseTest, break_dataset
+from integration_tests.helpers import ClickhouseTest, break_dataset
 from mindsdb_native import Predictor, ClickhouseDS
 
 
@@ -32,4 +32,4 @@ class TestMultitargetPrediction(ClickhouseTest):
             stop_training_in_x_seconds=3,
             use_gpu=False,
             advanced_args={'debug': True}
-        )
\ No newline at end of file
+        )
