@@ -35,8 +35,6 @@ class TestDataAnalyzer(unittest.TestCase):
 
         model_data = analyse_dataset(from_data=df)
         for col, col_data in model_data['data_analysis_v2'].items():
-            if col in ['columns','useable_input_columns']:
-                continue
             expected_type = test_column_types[col][0]
             expected_subtype = test_column_types[col][1]
             assert col_data['typing']['data_type'] == expected_type
@@ -152,8 +150,6 @@ class TestDataAnalyzer(unittest.TestCase):
             assert mock_function.called
 
         for col, col_data in model_data['data_analysis_v2'].items():
-            if col in ['columns','useable_input_columns']:
-                continue
             expected_type = test_column_types[col][0]
             expected_subtype = test_column_types[col][1]
             assert col_data['typing']['data_type'] == expected_type
